"""
utils.py is a file for storing utility classes exclusively used in the
TensorFlow/Keras ecosystem.

The module contains:

    - The 'TensorFlowDatasetConfigurer' class:
        Used for configuring Keras and transforming data into tf.data.Dataset
        objects ready to be used.
        
    - The 'LogsTracker' class:
        An class for storing/caching metrics on the fly.
        
    - The 'ProgressBar' class:
        A class for printing a progressbar.
"""

# Author: Mikael Frenette (mik.frenette@gmail.com)
# Created: September 2023

import numpy as np
import pandas as pd
import sys
import tensorflow as tf
from typing import Union

class TensorFlowDatasetConfigurer(object):
    """
    Author: Mikael Frenette
    
    DESCRIPTION:
    ---------------------------------------------------------------------------
    A class for configuring TensorFlow datasets.
    ---------------------------------------------------------------------------
    
    PARAMETERS
    ----------
    batch_size  : The batch size for the dataset.
    
    repeat      : The number of times to repeat the dataset. Default is 1.
    
    shuffle     : Whether to shuffle the dataset. Default is False.
    
    buffer_size : The buffer size for shuffling the dataset. Default is 32.
    
    prefetch    : The number of elements to prefetch in the dataset.
                  Default is 1.
    
    USAGE
    -----
    X, y = numpy arrays
    
    configurer = DatasetConfigurer(batch_size=5, repeat=1, prefetch=1,
                                   shuffle=True, buffer_size=32)
    dataset = configurer(X=X, y=y)
    ---------------------------------------------------------------------------
    """

    def __init__(self, 
                 batch_size: int = None,
                 repeat: int = 1,
                 shuffle: bool = False,
                 buffer_size: int = tf.data.AUTOTUNE,
                 prefetch: int = 1) -> None:
        
        self.batch_size = batch_size
        self.repeat = repeat
        self.shuffle = shuffle
        self.buffer_size = buffer_size
        self.prefetch = prefetch

    def __call__(self, 
                 X: Union[tuple[Union[np.ndarray, tf.Tensor],
                                Union[np.ndarray, tf.Tensor]],
                          np.ndarray, tf.Tensor],
                 y: Union[tuple[Union[np.ndarray, tf.Tensor],
                                Union[np.ndarray, tf.Tensor]],
                          np.ndarray, tf.Tensor] = None) -> tf.data.Dataset:
        """
        PARAMETERS
        ----------
        X : The input data.
            
        y : The target data. Default is None.

        RETURNS
            tf.data.Dataset: The configured TensorFlow dataset.
        """

        if not isinstance(X, tuple):
            X = (X)
        Xs = tf.data.Dataset.from_tensor_slices(X)
        
        if y is not None:
            if not isinstance(y, tuple):
                y = (y)
            ys = tf.data.Dataset.from_tensor_slices(y)
            dataset = tf.data.Dataset.zip((Xs, ys))
            del Xs, ys
        else:
            dataset = Xs
            del Xs

        dataset = dataset.repeat(self.repeat)
        dataset = dataset.prefetch(self.prefetch)  # Corrected prefetch argument
        
        if self.shuffle:
            dataset = dataset.shuffle(self.buffer_size)
            
        if self.batch_size is None:
            dataset = dataset.batch(len(X))
        else:
            dataset = dataset.batch(self.batch_size) 
            
        return dataset
    
    
    @staticmethod
    def from_default_config(X: Union[tuple[Union[np.ndarray, tf.Tensor],
                                       Union[np.ndarray, tf.Tensor]],
                                    np.ndarray, tf.Tensor],
                           y: Union[tuple[Union[np.ndarray, tf.Tensor],
                                      Union[np.ndarray, tf.Tensor]],
                                    np.ndarray, tf.Tensor] = None) -> tf.data.Dataset:
        """
        Create and configure a TensorFlow dataset.
    
        PARAMETERS
        ----------
        X : The input data.
        
        y : The target data. Default is None.
    
        RETURNS
            tf.data.Dataset: The configured TensorFlow dataset.
        """
        if not isinstance(X, tuple):
            X = (X)
        Xs = tf.data.Dataset.from_tensor_slices(X)
        
        if y is not None:
            if not isinstance(y, tuple):
                y = (y)
            ys = tf.data.Dataset.from_tensor_slices(y)
            dataset = tf.data.Dataset.zip((Xs, ys))
        else:
            dataset = Xs

        dataset = dataset.repeat(1)
        dataset = dataset.prefetch(tf.data.AUTOTUNE)  # Corrected prefetch argument
        dataset = dataset.shuffle(len(X))
        dataset = dataset.batch(32)

        return dataset


class LogsTracker(object):
    """
    DESCRIPTION
    ---------------------------------------------------------------------------
    The LogsTracker class is designed for the purpose of storing loss and
    metric values during training. This is achieved by maintaining a dictionary
    where each metric is associated with a list.

    Access to the log history can be obtained using the .history property.
    It's important to note that the reset_state() method must be called before
    utilizing it.
    ---------------------------------------------------------------------------
    """

    def __init__(self):
        self._logs = {}
        
    def __getitem__(self, metric:str):
        return self._logs[metric][-1]

    def __setitem__(self, metric:str, value:float):
        if metric not in self._logs.keys():
            self._logs[metric] = []
            
        if tf.is_tensor(value):
            self._logs[metric].append(value)
        
    def __len__(self):
        return len(self._logs)
    
    def last_log(self):
        last_log = {}
        for key, val in self._logs.items():
            last_log[key] = val[-1]
        return last_log
    
    def reset_state(self):
        self._logs = {}
        
    @property
    def logs(self):
        return self._logs
    
    @property
    def history(self):
        return pd.DataFrame.from_dict(self._logs)
    
    
class ProgressBar(object):
    """
    DESCRIPTION
    ---------------------------------------------------------------------------
    A custom progress bar for displaying training progress. This class allows
    for displaying a progress bar during training, indicating the progress of
    epochs and steps.
    ---------------------------------------------------------------------------

    PARAMETERS
    ----------
    total_epochs        : Total number of epochs.
    
    steps_per_epoch     : Total number of steps per epoch.
    
    number_of_trainings : Total number of training processes. Defaults to 1.
    
    length              : Length of the progress bar. Defaults to 20.
    
    fill                : Character used to fill the progress bar.
                          Defaults to '█'.
    """
    def __init__(self, 
                 name:str,
                 total_epochs:int,
                 steps_per_epoch:int,
                 number_of_trainings:int=1,
                 length:int=20, fill='█')->None:

        self.name = name
        self.length = length
        self.fill = fill
        self.total_epochs = total_epochs
        self.steps_per_epoch = steps_per_epoch
        self.number_of_trainings = number_of_trainings
        
    def __call__(self, prefix:str, epoch:int, step:int, logs:dict):
        
        """
        PARAMETERS
        ----------
        prefix : Prefix string displayed before the progress bar.
        
        epoch  : Current epoch.
        
        step   : the substep in the entire epoch. also called the batch size
                 step.
                 
        logs   : a dict containing the losses/metrics which will be printed out.
                 The dict must contain float values and the keys must be the 
                 desired name to be displayed.
                 
        RETURNS
            None.
        """
        epoch_progress = epoch + 1
        epoch_percent = round(100 * (epoch_progress / int(self.total_epochs)))

        filled_length = int(
            self.length * (epoch_progress - 1) // self.total_epochs) + 1
        bar = self.fill * filled_length + '-' * (self.length - filled_length)
            
        base_template = ["\r"
            f"{self.name} {prefix}",
            f"|{bar}| Step: {step}/{self.steps_per_epoch}",
            f"| Epoch: {epoch_progress}/{self.total_epochs} ({epoch_percent}%)",
        ]
        
        for name, val in logs.items():     
            metric_format = f"| {name}: {val:.4f}"   
            base_template.append(metric_format)
        
        progress = " ".join(base_template)
        sys.stdout.write(progress)
        sys.stdout.flush()


"""
trainers.py is a module designed to store training algorithms for keras models.
The classes defined in this module are exclusively useful within the
TensorFlow/Keras ecosystem.

The module contains:

    - The 'Trainer' class: A class for standard neural network training.
      
"""

# Author: Mikael Frenette (mik.frenette@gmail.com)
# Created: September 2023

import tensorflow as tf
from aionx.kerasnn import base
from typing import Union, List, Callable

class Trainer(base.NetworkTrainer):
    
    """
    Author: Mikael Frenette
    
    DESCRIPTION
    ---------------------------------------------------------------------------
    This class provides a custom yet standardized way of training a Keras model.
    It inherits from the parent class 'NetworkTrainer,' which offers the 
    standard training loop. Training steps and validation steps are defined
    here, performing the standard procedures. Please note that the
    '@tf.function' decorators are used by default when training the network
    in graph mode (not eagerly). For more details about training in graph mode,
    please refer to TensorFlow's documentation.
    ---------------------------------------------------------------------------
    
    PARAMETERS
    ----------
    optimizer : Optimizer used for training the model.
    
    loss      : Loss function(s) used for training the model.
    
    metrics   : Metric(s) used for evaluation during training. Defaults to None.
                Also, please note that using metrics when training in graph mode
                will raise an error.
    
    **kwargs  : Additional keyword arguments to pass to the parent class
                constructor.
    
    RETURNS
        None
    
    USAGE
    -----
    One can use a Keras models object without compiling it.
    
    model = create_model()  # returns a Keras model
    
    trainer = NetworkTrainer(optimizer=keras.optimizers.Adam(1e-3),
                             loss=keras.losses.MeanSquaredError(),
                             metrics=[keras.metrics.RootMeanSquaredError()])
    
    trainer.train(model, X, y, epochs=20, ...)
    """
    
    def __init___(self,
                 optimizer:Union[tf.keras.optimizers.Optimizer,
                                 List[tf.keras.optimizers.Optimizer]],
                 loss:Union[Callable, tf.keras.losses.Loss,
                            List[tf.keras.losses.Loss]],
                 metrics:Union[tf.keras.metrics.Metric,
                               List[tf.keras.metrics.Metric]]=None,
                 **kwargs)->None:
        
        super().__init__(
                     optimizer=optimizer,
                     loss=loss,
                     metrics=metrics,
                     **kwargs
                     )

    @tf.function
    def training_step(self, x, y):
        with tf.GradientTape() as tape: # auto diff
            pred = self.model(x, training=True)
            loss = self.loss_fn(y, pred)  # compute loss
        grads = tape.gradient(loss, self.model.trainable_weights) # compute gradients
        # back propagate
        self.optimizer.apply_gradients(zip(grads, self.model.trainable_weights))  

        for i, metric in enumerate(self.metrics):
            #update logger with metrics values
            metric.update_state(y, pred)
            self.logger["train_"+metric.name] = metric.result().numpy()
        return loss

    @tf.function
    def validation_step(self, x, y):
        pred = self.model(x, training=False)       
        val_loss = self.loss_fn(y, pred) # compute validation loss
        for i, metric in enumerate(self.metrics):
            metric.update_state(y, pred)
            #update logger with metrics values
            self.logger["val_"+metric.name] = metric.result().numpy()
        return val_loss
    
    def history(self):
        return self.logger.history
        



"""
 base.py file for storing parent classes useful entirely in the Tensorflow/
 keras Ecosystem.
 
 The module contains:
     
     the 'NetworkTrainer' base class. Used for training keras models only.
"""

# Author: Mikael Frenette (mik.frenette@gmail.com)
# Created: September 2023

import tensorflow as tf
import numpy as np 
from typing import Union, List, Callable

from aionx.kerasnn.utils import (
    TensorFlowDatasetConfigurer,
    LogsTracker,
    ProgressBar
    )

class NetworkTrainer(object):
    """
    Author: Mikael Frenette
    
    DESCRIPTION
    --------------------------------------------------------------------------
    Base class used for training a keras.models.Model. Its purpose is to mimic
    TensorFlow's model.compile() and model.fit() functionalities as closely
    as possible.
    
    The class is instantiated by specifying a keras.optimizers instance and
    a callable loss function with a signature similar to (y_true, y_pred).
    --------------------------------------------------------------------------
    
    PARAMETERS
    ----------
    optimizer : The network optimizer.
    
    loss      : The loss function used for training the network.
    
    metrics   : A list of metrics that will be tracked and printed during training.
                Default to None.
              
    n_trainings : Number of models that will be trained. Only affects the 
                  the progress bar. Default to 1.
              
    METHODS
    -------
    train : The train method begins by transforming input data into a tf.data.Dataset.
            It then iterates over the dataset for a specified number of epochs and
            applies a training and validation step as defined by the child class.
    
    PROPERTIES
    ----------
    One can access the training history logs by using the .history property
    on the trainer instance, which contains information tracked by LogsTracker.

    """
    
    def __init__(self,
                 optimizer:Union[tf.keras.optimizers.Optimizer,
                                 List[tf.keras.optimizers.Optimizer]],
                 loss:Union[Callable, tf.keras.losses.Loss,
                            List[tf.keras.losses.Loss]],
                 metrics:Union[tf.keras.metrics.Metric,
                               List[tf.keras.metrics.Metric]]=None,
                 )->None:
        
        self.optimizer = optimizer
        self.loss_fn = loss
        
        if isinstance(metrics, list):
            self.metrics = metrics
        elif metrics is None:
            self.metrics = {}
        else:
            self.metrics = [metrics]
            
        self.logger = LogsTracker()
    
        
    @property
    def history(self):
        return self.logger.history()

    def train(self,
              model:tf.keras.models.Model,
              X:Union[tuple[Union[np.ndarray, tf.Tensor],
                            Union[np.ndarray, tf.Tensor]],
                              np.ndarray, tf.Tensor],
              y:Union[tuple[Union[np.ndarray, tf.Tensor],
                            Union[np.ndarray, tf.Tensor]],
                              np.ndarray, tf.Tensor],
              epochs:int,
              shuffle:bool=False,
              validation_data:tuple[Union[np.ndarray, tf.Tensor],
                                    Union[np.ndarray, tf.Tensor]]=None,
              batch_size:int=32,
              verbose:int=1,
              early_stopping:tf.keras.callbacks.EarlyStopping=None,
              validation_batch_size:int=None,
              logger_description:str=None,
              )->None:
        
        """
        
        DESCRIPTION
        ----------------------------------------------------------------------
        The `train` method follows a structure similar to keras.fit and is
        designed for training a Keras model.
        ----------------------------------------------------------------------
        
        PARAMETERS
        ----------
        model   : The model to be trained. It does not need to be compiled.
        
        X       : The input data for the network.
        
        y       : The target data.
        
        epochs  : The number of training epochs for the model.
        
        shuffle : If set to True, only the training data will be shuffled.
                  Default to False.
                  
        validation_data : Validation data to be monitored during training.
                          Default to None
        
        batch_size      : The batch size for applying a gradient descent step. 
                          Default to 32.
                     
        verbose         : The verbosity level. A higher value prints more
                          logs during training. If set to 0, the progress bar
                          will be suppressed.
                  
        early_stopping  : Performs early stopping based on the parameters
                          of the Keras EarlyStopping callback. Default to None.
                        
        validation_batch_size : The batch size for validation steps. Default
                                to None.
        ----------------------------------------------------------------------

        """
        
        # add the model as attribute since we need it in other methods
        self.model = model
        
        # this needs to be set to false for the early stopping
        self.model.stop_training=False
        

        # get the shape input shape
        if isinstance(X, tuple):
            # look at first instance and compute the first and last dimension
            # parameters
            N, K = X[0].shape[0], X[0].shape[-1]
            # in the case where Xs in the tuple do not have the same shape,
            # an error will be raised by tensorflow while configuring the 
            # dataset
        else:
            N, K = X.shape[0], X.shape[-1]
            
        # training dataset configuration
        train_configurer = TensorFlowDatasetConfigurer(
            batch_size=batch_size,
            repeat=1, prefetch=tf.data.AUTOTUNE,
            shuffle=shuffle,
            buffer_size=N
            )
        train_set = train_configurer(X, y)
        
        if validation_data is not None:
            # if validation_data is provided, then configure the validation
            # dataset
            X_val, y_val = validation_data
            
            if validation_batch_size is None:
                validation_batch_size = X_val[0].shape[0] if isinstance(X_val, tuple) else X_val.shape[0]

            val_configurer = TensorFlowDatasetConfigurer(
                batch_size=validation_batch_size,
                repeat=1, prefetch=tf.data.AUTOTUNE,
                shuffle=shuffle,
                buffer_size=N
                )
            val_set = val_configurer(X_val, y_val)

        num_batches = tf.data.experimental.cardinality(train_set).numpy()
        
        progress_bar = ProgressBar(name=self.model.name,
                                   total_epochs=epochs,
                                   steps_per_epoch=num_batches)
        
        # earlystopping requires on_train_begin() initialization.
        # see keras.callbacks documentations
        if early_stopping is not None:
            early_stopping.model = self.model
            early_stopping.on_train_begin()
        
        # reset logger's state before training
        self.logger.reset_state()
        
        # main loop over epochs
        for epoch in range(epochs): 
            train_loss = 0.0 
            # a regular train step
            for train_step, (x_batch_train, y_batch_train) in enumerate(train_set):
                batch_loss = self.training_step(x_batch_train, y_batch_train)
                train_loss+= batch_loss
                train_loss = train_loss / (train_step + 1)
                self.logger["loss"] = train_loss
                
                if verbose>0: # print out progress_bar
                    progress_bar(f"{logger_description}",
                                 epoch, train_step+1, self.logger.last_log())
                                   
            if validation_data is not None:
                val_loss = 0.0
                # a regular validation step
                for val_step, (x_batch_val, y_batch_val) in enumerate(val_set):
                    batch_vloss = self.validation_step(x_batch_val, y_batch_val)   
                    val_loss+=batch_vloss
                    val_loss = val_loss / (val_step+1)
                self.logger["val_loss"] = val_loss
                
            for metric in self.metrics:
                metric.reset_state()
            
            # watch early stopping status.
            if early_stopping is not None:
                early_stopping.on_epoch_end(epoch, self.logger.last_log())
                if early_stopping.model.stop_training:
                    break # force training stop
        
    
                    
        
    



"""
aionx metrics.py is a file for storing metric classes.

The module contains:
    
    - The 'MeanSquaredError' class:
        A class for computing the mean squared error. 
        
    - The 'MeanAbsoluteError' class:
        A class for computing the mean absolute error.
        
    - The 'RootMeanSquaredError' class:
        A class for computing the root of the mean squared error.
        
    - The 'LogScore' class:
        A class for computing the Gaussian log score.
        
    - The 'NominalCoverage' class:
        A class for computing the norminal coverage.
"""

# Author: Mikael Frenette (mik.frenette@gmail.com)
# Created: September 2023

import numpy as np
from typing import List
from aionx import base

class MeanSquaredError(base.RegressionMetric):
    """
    DESCRIPTION
    ---------------------------------------------------------------------------
    Class to compute the Mean squared error between a provided true target
    (y_true) and predicted values (y_pred). Inherits from RegressionMetrics
    which controls for inputs formatting.
    ---------------------------------------------------------------------------
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __call__(self,
                 y_true:np.ndarray,
                 y_pred:np.ndarray) -> float:
        """
        PARAMETERS
        ----------
        y_true : True values.
        
        y_pred : Predicted value

        RETURNS
            float : the computed metric with the specified precision type.
        """
        super().__call__(y_true, y_pred)
        # Calculate the mean squared error
        metric = np.mean((y_true - y_pred) ** 2)
        return np.array(metric, dtype=self.precision)



class MeanAbsoluteError(base.RegressionMetric):
    """
    DESCRIPTION
    ---------------------------------------------------------------------------
    Class to compute the Mean absolute error between a provided true target
    (y_true) and predicted values (y_pred). Inherits from RegressionMetrics
    which controls for inputs formatting.
    ---------------------------------------------------------------------------
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __call__(self,
                 y_true:np.ndarray,
                 y_pred:np.ndarray) -> float:
        """
        PARAMETERS
        ----------
        y_true : True values.
        
        y_pred : Predicted value

        RETURNS
            float : the computed metric with the specified precision type.
        """
        super().__call__(y_true, y_pred)
        
        # Calculate the mean absolute error
        metric = np.mean(np.abs(y_true - y_pred))
        
        return np.array(metric, dtype=self.precision)


class RootMeanSquaredError(base.RegressionMetric):
    """
    DESCRIPTION
    ---------------------------------------------------------------------------
    Class to compute the Root Mean squared error between a provided true target
    (y_true) and predicted values (y_pred). Inherits from RegressionMetrics
    which controls for inputs formatting.
    ---------------------------------------------------------------------------
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __call__(self,
                 y_true:np.ndarray,
                 y_pred:np.ndarray) -> float:
        """
        PARAMETERS
        ----------
        y_true : True values.
        
        y_pred : Predicted value

        RETURNS
            float : the computed metric with the specified precision type.
        """
        super().__call__(y_true, y_pred)
        
        # Calculate the root mean squared error
        metric = np.sqrt(np.mean((y_true - y_pred) ** 2))
        
        return np.array(metric, dtype=self.precision)
    

class LogScore(base.DensityMetric):
    """
    DESCRIPTION
    ---------------------------------------------------------------------------
    Class to compute the Gaussian Log Score between a provided true target
    (y_true) and a list containing predictions for the mean and standard
    deviation (y_pred). Inherits from DensityMetrics which controls for inputs
    formatting.
    ---------------------------------------------------------------------------
    """
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
    def __call__(self,
                 y_true:np.ndarray,
                 y_pred:List[np.ndarray]) -> float:
        """
        PARAMETERS
        ----------
        y_true : True values.
        
        y_pred : predicted mean and standard deviationb values

        RETURNS
            float : the computed metric with the specified precision type.
        """
        y_true, y_pred = super().__call__(y_true, y_pred)
        
        # Convert inputs to numpy arrays for ease of calculation
        y, mu, sigma = y_true, np.squeeze(y_pred[0]), np.squeeze(y_pred[1])

        # Calculate the error between true values and mean predictions
        eps = y - mu

        # Calculate the log probability density function using the error and standard deviation
        log_pdf = -0.5 * np.log(2 * np.pi * sigma ** 2) - 0.5 * (eps ** 2) / ( (sigma ** 2) + self.tolerance )

        # Return the negative of the log likelihood as the log score
        return np.array(-np.mean(log_pdf), dtype=self.precision)



class NominalCoverage(base.DensityMetric):
    """
    DESCRIPTION
    ---------------------------------------------------------------------------
    Class to compute the Gaussian Log Score between a provided true target
    (y_true) and a list containing predictions for the mean and standard
    deviation (y_pred). Inherits from DensityMetrics which controls for inputs
    formatting.
    
    computes the ratio where the residuals fall into the standard deviation
    brackets
    
    if std == 1:
        computes eps +/- 1*sigma
        
    if std == 2:
        computes eps +/- 2*sigma
        
    ...        
    ---------------------------------------------------------------------------
    
    PARAMETERS
    ----------
    std : The number of standard deviations to use for the coverage
          interval. Defaults to 1.
    """
    def __init__(self, std:int=1, **kwargs):
        super().__init__(**kwargs)
        self.std = std

    def __call__(self,
                 y_true:np.ndarray,
                 y_pred:List[np.ndarray]) -> float:
        """
        PARAMETERS
        ----------
        y_true : True values.
        
        y_pred : predicted mean and standard deviationb values

        RETURNS
            float : the computed metric with the specified precision type.
        """
        y_true, y_pred = super().__call__(y_true, y_pred)
        y, mu, sigma = y_true, np.squeeze(y_pred[0]), np.squeeze(y_pred[1])

        # Compute the boolean masks for values outside the coverage interval
        above_interval = y - mu > (self.std * sigma)
        below_interval = y - mu < (-self.std * sigma)
        outside_interval = np.logical_or(above_interval, below_interval)

        # Calculate the number of values outside the interval
        not_covered = np.sum(outside_interval)

        # Calculate the coverage ratio
        coverage = 1 - (not_covered / len(y))

        return coverage






import numpy as np
import pandas as pd
import time
from typing import Union
from functools import wraps


def timeit(func):
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        elapsed_time = end_time - start_time
        hours, remainder = divmod(elapsed_time, 3600)
        minutes, seconds = divmod(remainder, 60)
        print(f" completed in {int(hours):02d}:{int(minutes):02d}:{int(seconds):02d}")
        return result
    return wrapper


def downstream(output_type:Union[np.ndarray,
                                 pd.Series,
                                 pd.DataFrame],
               dtype:str):
    """
    DESCRIPTION
    ---------------------------------------------------------------------------
    This decorator allows you to specify the desired downstream output type for
    a function. It automatically performs the conversion as needed to ensure
    the output matches the specified type.
    
    wrapper assumes 3 possible type for the function's output:
        {np.array, pd.DataFrame, pd.Series}.
    
    If the function is not one of these type, no transformation will be
    applied.
    ---------------------------------------------------------------------------
    
    PARAMETERS
    ----------
    output_type : The desired output type, which can be one of the following:
                  {np.ndarray, pd.Series, pd.DataFrame}.
                  
    dtype       : the output data type expected by pandas or numpy.
                  {int32, int64, float32, float64}.

    RETURNS
        A wrapped function with the specified output type.
    """
    
    if output_type not in [np.ndarray, pd.DataFrame, pd.Series]:
        raise ValueError("Specified output type must be one of " +
                         "{numpy.ndarray, pandas.DataFrame, pandas.Series}. " +
                         f"input received : {output_type}.")
    
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            output = func(*args, **kwargs)

            # if we want numpy array as output.
            if output_type == np.ndarray:
                if isinstance(output, pd.Series) or isinstance(output,
                                                               pd.DataFrame):
                    return output.values

            # If we want a pandas Serie as output
            elif output_type == pd.Series:
                if isinstance(output, np.ndarray):
                    return pd.Series(np.squeeze(output))
                elif isinstance(output, pd.DataFrame):
                    return output.squeeze()
              
            # If we want a pandas DataFrame as output
            elif output_type == pd.DataFrame:
                if isinstance(output, pd.Series):
                    return output.to_frame(output.name)
                elif isinstance(output, np.ndarray):
                    return pd.DataFrame(output)
                
            if dtype is not None:
                output = output.astype(dtype)

            return output # return output with the specified type.

        return wrapper

    return decorator


"""
aionx models.py is a file for storing model's classes.

The module contains:
    
    - The 'DensityHNN' class:
        A class implementation of the model proposed in
        'From Reactive to Proactive Volatility Modeling with
        Hemisphere Neural Networks' by Goulet-Coulombe (2023).
        
        
"""

# Author: Mikael Frenette (mik.frenette@gmail.com)
# Created: october 2023

import tensorflow as tf
from tensorflow import keras
from keras.models import Sequential, Model
from keras.initializers import RandomNormal, he_uniform
from keras.layers import Flatten, Dense, Input, Dropout
import numpy as np
from datetime import datetime
from typing import Union
import pandas as pd
from sklearn.metrics import mean_squared_error
from sklearn.linear_model import LinearRegression

from aionx.scalers import StandardScaler
from aionx.pipelines import WindowDataset
from aionx.kerasnn.trainers import Trainer
from aionx.utils import (add_trends, ExpandingWindowGenerator)
from aionx.bootstrap import TimeSeriesBlockBootstrap
from aionx.kerasnn.ensemble import DeepEnsemble, OutOfBagPredictor
from aionx.kerasnn.losses import GaussianLogLikelihood, InflationMSE

import warnings

    
class ScalingLayer(keras.layers.Layer):
    def __init__(self, beta, **kwargs):
        super().__init__(**kwargs)
        self.beta = beta
    def call(self, X):
        X = self.beta*X
        return X

class DensityHNN(object):
    """"
    Author : Mikael Frenette (mik.frenette@gmail.com)
    
    DESCRIPTION
    ---------------------------------------------------------------------------
    Initialize a Density Hemisphere Neural Network (DensityHNN) model.
    DensityHNN is an ensemble of neural network trained by minimizing the
    negative log likelihood in order to produce forecasts along with associated
    uncertainty estimates. 
    
                                  IMPORTANT
    ***************************************************************************

    THE 'DensityHNN' CLASS IS DESIGNED FOR SMALL DATASETS SUCH AS FRED-MD OR
    FRED-QD. DO NOT USE THIS CLASS WITH DATA SUBJECTED TO IREGULAR TIMESTAMPS
    SUCH AS STOCK MARKET DATA (WHICH HAVE HOLIDAYS), BECAUSE PANDAS CANNOT
    INFER FROM THEIR TIMESTAMP EASILY. ADDITIONALY, I WOULD RECOMMEND NOT USING
    THE CLASS ON LARGE DATASET SINCE THIS IMPLEMENTATION DOES NOT SUPPORT A
    MINIBATCH LEARNING AND LARGE BATCHES MAY QUICKY LEAD TO MEMORY ERROR.
    
    ***************************************************************************
    
    ---------------------------------------------------------------------------
                   
    PARAMETERS
    ----------
    target           : The target variable name.
    
    lags             : Number of lags used for each regressor provided.
    
    horizon          : The number of steps ahead to forecast. For example,
                       if horizon=1, the network will be trained to predict
                       y_{t+1} using y_{t-lags:t}.
    
    bootstraps       : Number of weak learners (estimators) to train. Their
                       forecasts will be averaged out. The default is 200.
    
    prior_bootstraps : Number of weak learners to use for the prior DNN. This
                       is used to compute the volatility emphasis parameter
                       (estimating a prior for the importance of volatility).
                       The default is 100.
    
    trends           : Number of trend regressors to include in the model.
                       The default is 100.
    
    sampling_rate    : The fraction of data to use for training each epoch.
                       The default is 0.8.
    
    block_size       : Size of blocks used for bootstrapping. The default is 8.
    
    epochs           : Number of training epochs. The default is 100.
    
    learning_rate    : Learning rate for training. The default is 1e-3.
    
    patience         : Number of epochs with no improvement after which
                       training will stop early. The default is 15.
    
    shuffle          : Whether to shuffle the training data before each epoch.
                       The default is False.
    
    verbose          : Verbosity mode (0, 1, or 2). The default is 1.
    
    STATIC METHODS
    --------------
    base_architecture : The base architecture method returns the architecture
                        used in the paper. One must force a volatility emphasis
                        parameter to use the model instead of estimating it
                        with a prior neural net.
    
    prior_dnn         : The prior_dnn method returns a Keras model
                        architecture. The model is used in an ensemble of
                        networks with identical architecture, and its 
                        out-of-bag error is used as volatility emphasis
                        for the DensityHNN.
    
    volatility_rescaling_algorithm : The rescaling function. It uses blocked
                                     out-of-bag residuals to recalibrate the
                                     ensemble's volatility estimation.
    
    METHODS
    -------
    run : The run method is used to instantiate, fit, and predict the data. One
          can also use expanding windows fitting, though this takes some time.
    
    USAGE
    -----
    HNN = DensityHNN(
        target="your_target_name",
        lags=6,
        horizon=1,
        bootstraps=100,
        prior_bootstraps=50,
    )
    
    results = HNN.run(your_data,
                      train_start="1961-01-01",
                      train_end="2006-12-01",
                      expanding=False)
    
    dnn_pred = results["dnn_forecast"]
    HNN_preds = results["hnn_forecast"]
    vol_emphasis = results["volatility_emphasis"]
    """

    def __init__(self,
                 target:str,
                 lags:int,
                 horizon:int,
                 bootstraps:int=200,
                 prior_bootstraps:int=100,
                 trends:int=100,
                 sampling_rate:float=0.8,
                 block_size:int=8,
                 epochs:int=100,
                 learning_rate:float=1e-3,
                 patience:int=15,
                 verbose:int=1,
                 )->None:
        
        # accessible attributes
        self.target = target
        self.lags = lags
        self.horizon = horizon
        self.bootstraps = bootstraps
        self.prior_bootstraps = prior_bootstraps
        self.trends = trends
        self.epochs=epochs
        self.sampling_rate = sampling_rate
        self.block_size = block_size
        self.learning_rate = learning_rate
        self.patience = patience
        self.verbose = verbose

        self.__pipeline = WindowDataset(
            in_steps = self.lags, out_steps = 1, horizon = self.horizon,
            targets = [self.target], sampler = None
            )
        
        self.__pred_pipeline = WindowDataset(
            in_steps = self.lags, out_steps = 0, horizon = 0,
            targets = [self.target], sampler = None
            )
        
    @staticmethod
    def base_architecture(
            input_shape:tuple,
            hemisphere_outputs:int=1,
            vol_emphasis:float=1.0
            )->Model:
        
        """
        PARAMETERS
        ----------
        input_shape        : format expected by keras models input.
        
        hemisphere_outputs : the number of output units for each hemisphere.
                             Default to 1.
                             
        vol_emphasis       : the vol emphasis parameter. must be between
                             0.01 and 1.0. Default to 1.0. If above 1.0, 
                             value will automatically be set to 0.99.
        
        RETURNS
            keras.models.Model
            
        """
        
        #The architecture is as following:
            
        #    - The network has 2 seperate hemispheres:
        #        One that models the conditional mean (hm).
        #        One that models the conditional std (hv).
        
        #    - Both hemispheres share a common block of layers at the entrance
        #      of the network (hz).
            
        #               -- hm -> mu
        #    X -> -- hz 
        #               -- hv -> sigma
                   
        if (hemisphere_outputs <=0):
            raise ValueError(
                "Each hemisphere must have atleast one output unit."+
                f"Number received  {hemisphere_outputs}."
                )
        
        if not (0.01 <= vol_emphasis <= 1.0):
            warnings.warn(
                "vol_emphasis input must be between 0.01 and 1.0 "+
                f"input received {vol_emphasis:.4f}. vol_emphasis will be set"+
                " to 0.99."
                )
            vol_emphasis = 0.99
            
        inputs = Input(shape=input_shape)
    
        shared = Sequential([
            Flatten(),
            Dense(units=400,
                activation="relu",
                kernel_initializer=RandomNormal(mean=0, stddev=0.03),
                ),
            Dropout(0.2),
            Dense(units=400,
                activation="relu",
                kernel_initializer=RandomNormal(mean=0, stddev=0.03),
                ),
            keras.layers.Dropout(0.2),
        ], name="shared_hemisphere")
        
        mean = Sequential([
            Dense(units=400,
                activation="relu",
                kernel_initializer=RandomNormal(mean=0, stddev=0.03),
                ),
            Dropout(0.2),
            Dense(units=400,
                activation="relu",
                kernel_initializer=RandomNormal(mean=0, stddev=0.03),
                ),
            Dropout(0.2),
            Dense(hemisphere_outputs, activation='linear'),
        ], name="conditional_mean_hemisphere")
        
        vol = Sequential([
            Dense(units=400,
                activation="relu",
                kernel_initializer=RandomNormal(mean=0, stddev=0.03),
                ),
            Dropout(0.2),
            Dense(units=400,
                activation="relu",
                kernel_initializer=RandomNormal(mean=0, stddev=0.03),
                ),
            Dropout(0.2),
            Dense(hemisphere_outputs, activation='softplus'),
        ], name="conditional_vol_hemisphere")
    
        z = shared(inputs)
        mean_output = mean(z)
        vol_output = vol(z)
        
        vol_output = tf.math.scalar_mul(
            vol_emphasis, tf.math.divide(vol_output,
                                         tf.reduce_mean(vol_output))
            )
    
        output = tf.concat([mean_output, vol_output], axis=-1)
    
        HNN = Model(inputs=[inputs], outputs=[output], name="density_hnn")
        
        HNN.compile(loss=GaussianLogLikelihood(),
                    optimizer=keras.optimizers.Adam(1e-3))
        
        return HNN
    
    @staticmethod
    def prior_dnn_architecture(input_shape:tuple)->Model:
        
        """
        PARAMETERS
        ----------
        input_shape : format expected by keras models input.
        
        RETURNS
            keras.models.Model.
        """
        
        inputs = Input(shape=input_shape)
        DNN = Sequential([
            Flatten(),
            Dense(units=400,
                  activation="tanh",
                  kernel_initializer=RandomNormal(mean=0, stddev=0.03)),
            Dropout(0.2),
            Dense(units=400,
                  activation="tanh",
                  kernel_initializer=RandomNormal(mean=0, stddev=0.03)),
            Dropout(0.2),
            Dense(units=400,
                  activation="tanh",
                  kernel_initializer=RandomNormal(mean=0, stddev=0.03)),
            Dropout(0.2),
            Dense(units=400,
                  activation="tanh",
                  kernel_initializer=RandomNormal(mean=0, stddev=0.03)),
            Dropout(0.2),
            Dense(1, activation="linear")
        ], name="dnn_block")
        
        output = DNN(inputs)
        
        DNN = Model(inputs=[inputs], outputs=[output], name="prior_dnn")
        
        DNN.compile(loss="MSE", optimizer=keras.optimizers.Adam(1e-3)) 
        
        return DNN
    
    @staticmethod
    def volatility_rescaling_algorithm(y_true:pd.Series,
                                     y_pred:pd.Series,
                                     vol_pred:pd.Series,
                                     oob_start:Union[str, datetime],
                                     oob_end:Union[str, datetime])->pd.Series:
        """
        PARAMETERS
        ----------
        y_true    : the true values of the target variable.
        
        y_pred    : the ensemble prediction of the conditional mean.
                    predictions on the training set must be out-of-bag.
                    
        vol_pred  : the ensemble prediction conditional volatility.
                    predictions on the training set must be out-of-bag.
                    
        oob_start : the start index of the out-of-bag (OOB) samples.
        
        oob_end   : the end index of the out-of-bag (OOB) samples.
    
        RETURNS
            VolatilityRescaled : the rescaled predicted volatility.
    
        """
    
        # concatenate the true and predicted values and drop any missing data
        merged = pd.concat([y_true, y_pred], axis=1).dropna()
    
        # calculate the model residuals
        HNN_residuals = merged.iloc[:, 0] - merged.iloc[:, 1]
    
        # concatenate the squared residuals and estimated volatility squared,
        # and only keep the
        # forecast on the train set (i.e., OOB)
        merged_oob = pd.concat([HNN_residuals**2,
                                vol_pred**2], axis=1).loc[oob_start:oob_end]
    
        # use the squared estimated volatility as the input (X) for the
        # linear regression
        X_train = merged_oob.iloc[:, 1].values.reshape(-1, 1)
    
        # use the squared residuals as the output (y) for the linear regression
        y_train = merged_oob.iloc[:, 0].values.reshape(-1, 1)
    
        # create a linear regression object and fit the training data
        linreg = LinearRegression()
        linreg.fit(np.log(X_train), np.log(y_train))
    
        # use the full predicted volatility as the input for the linear
        # regression
        full_X = np.log(vol_pred**2)
    
        # predict the log of the residuals based on the log of the estimated
        # volatility
        projection = linreg.predict(full_X.values.reshape(-1, 1))
    
        # store the predictions in a dataframe
        pred = pd.DataFrame(projection, columns=["vol"], index=y_pred.index)
    
        # merge the log squared errors of the model and the predictions and
        # drop any missing data
        merged2 = pd.concat([np.log(merged_oob.iloc[:, 0]), pred],
                            axis=1).dropna()
    
        # calculate the residuals between the two merged dataframes
        res = merged2.iloc[:, 0].values - merged2.iloc[:, 1].values
    
        # calculate the scaler by taking the square root of the mean
        # exponential of the residuals
        scaler = np.sqrt(np.mean(np.exp(res)))
    
        # adjust the predicted volatility based on the scaler
        VolatilityRescaled = np.sqrt(np.exp(pred)) * scaler
    
        return VolatilityRescaled
        
    def run(self,
              data:Union[pd.DataFrame, pd.Series],
              train_start = Union[str, datetime],
              train_end = Union[str, datetime],
              expanding:bool=False,
              expanding_start:Union[str, datetime]=None,
              expanding_steps:int=8,
              last_expanding_window:Union[str, datetime]=None
              )->dict:
        
        """
        PARAMETERS
        ----------
        data            : The input data to be used.
        
        train_start     : The starting date or timestamp for the training
                          data.
        
        train_end       : The ending date or timestamp for the training
                          data.
        
        expanding       : If True, use an expanding training window.
                           Default is False.
                    
        expanding_start : The starting date or timestamp for the expanding
                          training window.
                          
        expanding_steps : The number of steps to expand the training window.
                          Only relevant when 'expanding' is True.
                          Default is 8.
                          
        last_expanding_window : The end date or timestamp for the last
                                expanding training window. Only relevant
                                when 'expanding' is True.

        RETURNS
            results : a python dictionary the ensemble forecast for the
                      prior_dnn and the hnn. Also, the dictionary contains the
                      estimated volatility emphasis parameter.
        """

        # making sure the data respects the index constraints.
        data = self._validate_dataset(data)
                  
        if expanding: # performs __static_fit() for each expanding window.
            windows = ExpandingWindowGenerator(data,
                                              expanding_start=expanding_start,
                                              last_window=last_expanding_window,
                                              timesteps=expanding_steps, verbose=self.verbose)
            pred_idx = pd.date_range(
                data.index[0+self.lags+self.horizon-1],
                periods=len(data)-self.lags+1,
                freq=pd.infer_freq(data.index)
            )
            dnn_rolling_outputs = {}
            hnn_rolling_outputs = {}
            vol_emphasis        = {}
            
            # main estimation loop.
            for step, train, poos in windows:
                window_results = self._static_fit(data,
                                                  train_start = train.index[0],
                                                  train_end = train.index[-1])   
                dnn_rolling_outputs[f"{poos.index[0]}"] = window_results[
                    "dnn_forecast"]
                hnn_rolling_outputs[f"{poos.index[0]}"] = window_results[
                    "hnn_forecast"]
                vol_emphasis[f"{poos.index[0]}"] = window_results[
                    "volatility_emphasis"]
                
            dnn_forecast = pd.DataFrame(columns=["dnn_forecast"],
                            index=pred_idx)

            for i, (step, output) in enumerate(dnn_rolling_outputs.items()):
                if i == 0:
                    dnn_forecast.loc[
                        :train_end, "dnn_forecast"] = output.loc[
                            :train_end, "forecast"]             
                dnn_forecast.loc[
                    step:, "dnn_forecast"] = output.loc[step:, "forecast"]
            dnn_forecast = dnn_forecast.astype("float32")
                
            
            hnn_forecast = pd.DataFrame(columns=["conditional_mean",
                                                 "conditional_vol"],
                                        index=pred_idx) 
            for i, (step, output) in enumerate(hnn_rolling_outputs.items()):
                if i == 0:
                    hnn_forecast.loc[
                        :train_end, "conditional_mean"] = output.loc[
                            :train_end, "conditional_mean"]
                    hnn_forecast.loc[
                        :train_end, "conditional_vol"] = output.loc[
                            :train_end, "conditional_vol"]
                    
                hnn_forecast.loc[
                    step:, "conditional_mean"] = output.loc[
                        step:, "conditional_mean"]
                hnn_forecast.loc[
                    step:, "conditional_vol"] = output.loc[
                        step:, "conditional_vol"]
                        
            hnn_forecast = hnn_forecast.astype("float32")
                        
            # to keep track of the volatility emphasis across each expanding
            # window.
            volatility_emphasis = pd.DataFrame.from_dict(vol_emphasis,
                                                         orient="index")
            volatility_emphasis.columns = ["volatility_emphasis"]
            volatility_emphasis.index.name = "date"
            # the resulting dict.
            results = {"hnn_forecast":hnn_forecast,
                       "dnn_forecast":dnn_forecast,
                       "volatility_emphasis":volatility_emphasis}
        else:
            results = self._static_fit(data, train_start=train_start, 
                                      train_end=train_end)
        return results
    
    
    
    def _static_fit(self, data:Union[pd.DataFrame, pd.Series],
                   train_start:Union[str, datetime],
                   train_end:Union[str, datetime]):

        # adding trends to the data
        data = add_trends(data, self.trends)
                       
        #instantiate a scaler
        scaler = StandardScaler()
                       
        # get the statistical parameters needed on the train set (mean, std)
        scaler.get_stats(data.loc[train_start:train_end])
                       
        # scale using the computed mean and std
        full_scaled = scaler.scale(data)
                       
        # split
        train_scaled = full_scaled.loc[train_start:train_end]
        # create data ready to be used by the models
        X_train, y_train = self.__pipeline(train_scaled)
        X_full, y_full = self.__pred_pipeline(full_scaled)
        input_shape=(self.lags, X_train.shape[-1]) # defining input shapes.

        # instantiate sampler
        self.__sampler = TimeSeriesBlockBootstrap(
            X=X_train, y=y_train,
            sampling_rate=self.sampling_rate,
            block_size=self.block_size,
            replace=True
        )
    
        pred_idx = pd.date_range(
            data.index[0+self.lags+self.horizon-1],
            periods=len(data)-self.lags+1,
            freq=pd.infer_freq(data.index)
            )
        tf.keras.backend.clear_session() # this is important so we do not
        # get overlapping graphs or tensorflow session during estimations.

        # only one tensorflow callback is used for the model.
        callbacks = [keras.callbacks.EarlyStopping(
            monitor="val_loss", patience=self.patience)]

        # custom trainer
        dnn_trainer = Trainer(
            optimizer = keras.optimizers.Adam(self.learning_rate),
            loss      = keras.losses.MeanSquaredError(),
            metrics   = None,
        )
                       
        # instantiate and fit
        prior_dnn = DensityHNN.prior_dnn_architecture(input_shape)
        ensemble_dnn = DeepEnsemble( # wrap into a deep ensemble.
            n_estimators=self.prior_bootstraps,
            network = prior_dnn,
            trainer = dnn_trainer,
            sampler = self.__sampler
            )
        ensemble_dnn.fit(
            X_train, y_train, epochs=self.epochs,
            early_stopping=callbacks[0],
            batch_size=None,
            validation_batch_size=None,
            verbose=self.verbose
            )               
        # By default, the predict method will return all estimators forecasts.
        dnn_prediction = ensemble_dnn.predict(X_full,
                                   batch_size=X_full.shape[0],
                                   verbose=0)
        # gather and create bluprints for out-of-bag forecasts.
        oob_idx = ensemble_dnn.oob_indices
        oob_predictor = OutOfBagPredictor(oob_idx)

        # oob and oos forecasts
        dnn_oob = oob_predictor(np.copy(dnn_prediction[:len(y_train)]),
                                sampling_rate=self.sampling_rate)
        dnn_oos = np.mean(np.copy(dnn_prediction[len(y_train):]), axis=-1)
        dnn_forecast = np.concatenate([dnn_oob, dnn_oos], axis=0)
    
        # compute volatility emphasis using the MSE on the out-of-bag forecast.
        volatility_emphasis = mean_squared_error(dnn_oob, np.squeeze(y_train))
        
        dnn_pred = scaler.unscale(
            pd.DataFrame(dnn_forecast, columns=["forecast"], index=pred_idx),
            target=self.target, use_mean=True
        )
                       
        tf.keras.backend.clear_session()
        # instantiate and fit
        densityhnn = DensityHNN.base_architecture(
            input_shape, hemisphere_outputs=1, 
            vol_emphasis=tf.constant(volatility_emphasis, dtype=tf.float32)
            )
        # custom trainer
        hnn_trainer = Trainer(
            optimizer = keras.optimizers.Adam(self.learning_rate),
            loss      = GaussianLogLikelihood(),
            metrics   = None,
        )
        self.ensemble_hnn = DeepEnsemble( # wrap in ensemble.
            n_estimators=self.bootstraps,
            network = densityhnn,
            trainer = hnn_trainer,
            sampler = self.__sampler
        )
        self.ensemble_hnn.fit(
            X_train, y_train, epochs=self.epochs,
            early_stopping=callbacks[0],
            batch_size=None,
            validation_batch_size=None,
            verbose=self.verbose
            )
        # make prediction
        mean_preds, vol_preds = self.ensemble_hnn.predict(
            X_full, batch_size=X_full.shape[0], verbose=0
        )
        # gather and create bluprints for out-of-bag forecasts.
        oob_idx = self.ensemble_hnn.oob_indices
        oob_predictor = OutOfBagPredictor(oob_idx)
        hnn_mean_oob = oob_predictor(np.copy(mean_preds[:len(y_train)]),
                                     sampling_rate=self.sampling_rate)        
        hnn_mean_oos = np.mean(np.copy(mean_preds[len(y_train):]), axis=-1)
        hnn_mean = np.concatenate([hnn_mean_oob, hnn_mean_oos], axis=0)
    
        hnn_mean = scaler.unscale(
            pd.DataFrame(hnn_mean, columns=["conditional_mean"], index=pred_idx),
            self.target, use_mean=True
        )
        hnn_vol_oob = oob_predictor(np.copy(vol_preds[:len(y_train)]),
                                    sampling_rate=self.sampling_rate)
        hnn_vol_oos = np.mean(np.copy(vol_preds[len(y_train):]), axis=-1)
        hnn_vol = np.concatenate([hnn_vol_oob, hnn_vol_oos], axis=0)
        hnn_vol = scaler.unscale(
            pd.DataFrame(hnn_vol, columns=["conditional_vol"], index=pred_idx),
            self.target, use_mean=False
        )
        hnn_preds = pd.concat([hnn_mean, hnn_vol], axis=1)
        hnn_preds.columns =["conditional_mean", "conditional_vol"]
        hnn_preds.index = pred_idx
                       
        # rescaling the volatility
        hnn_preds["conditional_vol"] = DensityHNN.volatility_rescaling_algorithm(
                                y_true    = data[[self.target]],
                                y_pred    = hnn_preds["conditional_mean"],
                                vol_pred  = hnn_preds["conditional_vol"],
                                oob_start = train_start,
                                oob_end   = train_end)
        
        return {"hnn_forecast":hnn_preds,
                "dnn_forecast":dnn_pred,
                "dnn_bootstraps":dnn_prediction,
                "hnn_bootstraps":{"conditional_mean":mean_preds,
                                  "conditional_vol":vol_preds},
                "volatility_emphasis":volatility_emphasis}
    
    def _validate_dataset(self, data):
        try: 
            pd.infer_freq(data.index)
        except TypeError:
            raise ValueError("Pandas cannot infer frequency from the provided "+
                             "data's index. Please make sure that your data "+
                             "has a valid datetime index. See panda's " +
                             "documentation for more information.")
        else:
            return data
    
class InflationHNN(object):
    """
    PARAMETERS
    ----------
    target           : The target variable name.
    
    lags             : Number of lags used for each regressor provided.
    
    horizon          : The number of steps ahead to forecast. For example,
                       if horizon=1, the network will be trained to predict
                       y_{t+1} using y_{t-lags:t}.
    
    bootstraps       : Number of weak learners (estimators) to train. Their
                       forecasts will be averaged out. The default is 200.
    
    prior_bootstraps : Number of weak learners to use for the prior DNN. This
                       is used to compute the volatility emphasis parameter
                       (estimating a prior for the importance of volatility).
                       The default is 100.
    
    trends           : Number of trend regressors to include in the model.
                       The default is 100.
    
    sampling_rate    : The fraction of data to use for training each epoch.
                       The default is 0.8.
    
    block_size       : Size of blocks used for bootstrapping. The default is 8.
    
    epochs           : Number of training epochs. The default is 100.
    
    learning_rate    : Learning rate for training. The default is 1e-3.
    
    patience         : Number of epochs with no improvement after which
                       training will stop early. The default is 15.
    
    shuffle          : Whether to shuffle the training data before each epoch.
                       The default is False.
    
    verbose          : Verbosity mode (0, 1, or 2). The default is 1.
    """
    def __init__(self,
                 target:str,
                 lags:int,
                 horizon:int,
                 bootstraps:int=200,
                 prior_bootstraps:int=100,
                 trends:int=100,
                 sampling_rate:float=0.8,
                 block_size:int=8,
                 epochs:int=100,
                 learning_rate:float=1e-3,
                 patience:int=15,
                 verbose:int=1,
                 )->None:
        
        # accessible attributes
        self.target = target
        self.lags = lags
        self.horizon = horizon
        self.bootstraps = bootstraps
        self.prior_bootstraps = prior_bootstraps
        self.trends = trends
        self.epochs=epochs
        self.sampling_rate = sampling_rate
        self.block_size = block_size
        self.learning_rate = learning_rate
        self.patience = patience
        self.verbose = verbose
    
    @staticmethod
    def base_architecture(
        activity_input_shape,
        prices_input_shape,
        LR_expectations_input_shape,
        SR_expectations_input_shape
        ):
        
        activity_beta = np.sqrt(activity_input_shape[-1])
        prices_beta = np.sqrt(prices_input_shape[-1])
        SR_beta = np.sqrt(LR_expectations_input_shape[-1])
        LR_beta = np.sqrt(SR_expectations_input_shape[-1])
        
        activity_inputs = keras.layers.Input(shape=(activity_input_shape))
        prices_inputs = keras.layers.Input(shape=(prices_input_shape))
        LR_inputs = keras.layers.Input(shape=(LR_expectations_input_shape))
        SR_inputs = keras.layers.Input(shape=(SR_expectations_input_shape))

        activity_hemisphere = keras.models.Sequential([
            ScalingLayer(beta=1/activity_beta),
            Flatten(),
            Dense(units=400, activation="relu", kernel_initializer=he_uniform()),
            Dropout(0.2),
            Dense(units=400, activation="relu", kernel_initializer=he_uniform()),
            Dropout(0.2),
            Dense(1)
        ])
        prices_hemisphere = keras.models.Sequential([
            ScalingLayer(beta=1/prices_beta),
            Flatten(),
            Dense(units=400, activation="relu", kernel_initializer=he_uniform()),
            Dropout(0.2),
            Dense(units=400, activation="relu", kernel_initializer=he_uniform()),
            Dropout(0.2),
            Dense(1)
        ])
        SR_hemisphere = keras.models.Sequential([
            ScalingLayer(beta=1/SR_beta),
            Flatten(),
            Dense(units=400, activation="relu", kernel_initializer=he_uniform()),
            Dropout(0.2),
            Dense(units=400, activation="relu", kernel_initializer=he_uniform()),
            Dropout(0.2),
            Dense(1)
        ])
        LR_hemisphere = keras.models.Sequential([
            ScalingLayer(beta=1/LR_beta),
            Flatten(),
            Dense(units=400, activation="relu", kernel_initializer=he_uniform()),
            Dropout(0.2),
            Dense(units=400, activation="relu", kernel_initializer=he_uniform()),
            Dropout(0.2),
            Dense(1)
        ])

        activity_output = activity_hemisphere(activity_inputs)
        prices_output = prices_hemisphere(prices_inputs)
        SR_output = SR_hemisphere(SR_inputs)
        LR_output = LR_hemisphere(LR_inputs)

        output = tf.concat([activity_output, prices_output, SR_output, LR_output], axis=-1)

        model = keras.models.Model(inputs=[activity_inputs, prices_inputs, LR_inputs, SR_inputs],
                                  outputs=output)

        return model

        
    def run(self,
              data:Union[pd.DataFrame, pd.Series],
              train_start = Union[str, datetime],
              train_end = Union[str, datetime],
              expanding:bool=False,
              expanding_start:Union[str, datetime]=None,
              expanding_steps:int=8,
              last_expanding_window:Union[str, datetime]=None
              )->dict:
        
        """
        PARAMETERS
        ----------
        data            : The input data to be used.
        
        train_start     : The starting date or timestamp for the training
                          data.
        
        train_end       : The ending date or timestamp for the training
                          data.
        
        expanding       : If True, use an expanding training window.
                           Default is False.
                    
        expanding_start : The starting date or timestamp for the expanding
                          training window.
                          
        expanding_steps : The number of steps to expand the training window.
                          Only relevant when 'expanding' is True.
                          Default is 8.
                          
        last_expanding_window : The end date or timestamp for the last
                                expanding training window. Only relevant
                                when 'expanding' is True.

        """

        # making sure the data respects the index constraints.
        data = self._validate_dataset(data)

        if expanding: # performs __static_fit() for each expanding window.
            
            dataset = (
                data["real_activity_input"],
                data["prices_input"],
                data["sr_expectations_input"],
                data["lr_expectations_input"]
            )

            windows = ExpandingWindowGenerator(dataset,
                                              expanding_start=expanding_start,
                                              last_window=last_expanding_window,
                                              timesteps=expanding_steps,
                                              verbose=self.verbose)
            pred_idx = pd.date_range(
                dataset[0].index[0+self.lags+self.horizon-1],
                periods=len(dataset[0])-self.lags+1,
                freq=pd.infer_freq(dataset[0].index)
            )

            hnn_rolling_outputs = {}
            
            # main estimation loop.
            for step, train, poos in windows:

                train_start_idx, train_end_idx = train[0].index[0], train[0].index[-1]
                test_start_idx, test_end_idx = poos[0].index[0], poos[0].index[-1]
                
                window_results = self._static_fit(data,
                                                  train_start = train_start_idx,
                                                  train_end = train_end_idx)   
                hnn_rolling_outputs[f"{test_start_idx}"] = window_results["components"]
                
            hnn_forecasts = pd.DataFrame(columns=["real_activity",
                                                 "prices", "short_run_expectations",
                                                 "long_run_expectations"],
                                        index=pred_idx) 
            for i, (step, output) in enumerate(hnn_rolling_outputs.items()):
                if i == 0:
                    hnn_forecasts.loc[
                        :train_end, "real_activity"] = output.loc[
                            :train_end, "real_activity"]
                    hnn_forecasts.loc[
                        :train_end, "prices"] = output.loc[
                            :train_end, "prices"]
                    hnn_forecasts.loc[
                        :train_end, "short_run_expectations"] = output.loc[
                            :train_end, "short_run_expectations"]
                    hnn_forecasts.loc[
                        :train_end, "long_run_expectations"] = output.loc[
                            :train_end, "long_run_expectations"]
                    
                hnn_forecasts.loc[
                    step:, "real_activity"] = output.loc[
                        step:, "real_activity"]
                hnn_forecasts.loc[
                    step:, "prices"] = output.loc[
                        step:, "prices"]
                hnn_forecasts.loc[
                    step:, "short_run_expectations"] = output.loc[
                        step:, "short_run_expectations"]
                hnn_forecasts.loc[
                    step:, "long_run_expectations"] = output.loc[
                        step:, "long_run_expectations"]
                
            hnn_forecasts = hnn_forecasts.astype("float32")
                    
            # the resulting dict.
            results = {"components":hnn_forecasts}
        else:
            results = self._static_fit(data, train_start=train_start, 
                                      train_end=train_end)
        return results
    
    
    
    def _static_fit(self,
                   data:Union[pd.DataFrame, pd.Series],
                   train_start:Union[str, datetime],
                   train_end:Union[str, datetime]
                    ):

        scalers, input_dict = [], {}
        Xs_train, ys_train, Xs_full, ys_full = [], [], [], []
        for hemisphere, inputs in data.items():
            scaler = StandardScaler()
            scaler.get_stats(inputs.loc[train_start:train_end])
            full_scaled = scaler.scale(inputs)
            train_scaled = full_scaled.loc[:train_end] 

            if self.target in inputs.columns:
                pipe = WindowDataset(targets=[self.target], in_steps=self.lags, horizon=self.horizon,
                                     sampler=None)
                X_train, y_train = pipe(train_scaled)
                full_pipe = WindowDataset(targets=None, in_steps=self.lags, horizon=0, out_steps=0,
                                          sampler=None)
                X_full, _ = full_pipe(full_scaled)
                Xs_train.append(X_train)
                ys_train.append(y_train)
                Xs_full.append(X_full)

            else:
                pipe = WindowDataset(targets=None, in_steps=self.lags, horizon=self.horizon,
                                     sampler=None)
                X_train, _ = pipe(train_scaled)
                full_pipe = WindowDataset(targets=None, in_steps=self.lags, horizon=0,
                                          out_steps=0, sampler=None)
                X_full, _ = full_pipe(full_scaled)
                Xs_train.append(X_train)
                Xs_full.append(X_full)
                
            input_dict[hemisphere] = X_train
            scalers.append(scaler)

        X_train, y_train = tuple(Xs_train), tuple(ys_train)
        X_full = tuple(Xs_full)
        # instantiate sampler
        self.__sampler = TimeSeriesBlockBootstrap(
            X=X_train, y=y_train,
            sampling_rate=self.sampling_rate,
            block_size=self.block_size,
            replace=True
        )

        def build_model():
            # build prior fully connected network
            model = InflationHNN.base_architecture(
                activity_input_shape = input_dict["real_activity_input"].shape[1:], 
                prices_input_shape = input_dict["prices_input"].shape[1:], 
                LR_expectations_input_shape = input_dict["sr_expectations_input"].shape[1:], 
                SR_expectations_input_shape = input_dict["lr_expectations_input"].shape[1:]
            )
            return model
        
        def build_trainer():
            trainer = Trainer(
            optimizer = keras.optimizers.Adam(self.learning_rate),
            loss      = InflationMSE(),
            metrics   = None,
            )
            return trainer 
            
        
        pred_idx = pd.date_range(
            data["prices_input"].index[0+self.lags+self.horizon-1],
            periods=len(data["prices_input"])-self.lags+1,
            freq=pd.infer_freq(data["prices_input"].index)
            )
        tf.keras.backend.clear_session() # this is important so we do not
        # get overlapping graphs or tensorflow session during estimations.

        # only one tensorflow callback is used for the model.
        callbacks = [keras.callbacks.EarlyStopping(
            monitor="val_loss", patience=self.patience)]

      
        # instantiate and fit
        ensemble_model = DeepEnsemble.from_function( # wrap into a deep ensemble.
            n_estimators=self.bootstraps,
            func = build_model,
            trainer_func = build_trainer,
            sampler = self.__sampler
            )
        ensemble_model.fit(
            X_train, y_train, epochs=self.epochs,
            early_stopping=callbacks[0],
            batch_size=len(X_train[0]),
            validation_batch_size=None,
            verbose=self.verbose
            )               
        # By default, the predict method will return all estimators forecasts.
        prediction = ensemble_model.predict(X_full, verbose=0)
        # gather and create bluprints for out-of-bag forecasts.
        oob_idx = ensemble_model.oob_indices
        oob_predictor = OutOfBagPredictor(oob_idx)
        
        forecasts = pd.DataFrame(index=pred_idx)
        
        for i, keys in enumerate(
            ("real_activity", "prices", "short_run_expectations", "long_run_expectations")
        ):
            
            oob_forecast = oob_predictor( np.copy(prediction[i][:len(y_train[0])]) )
            oos_forecast = np.mean( np.copy(prediction[i][len(y_train[0]):]), axis=-1)

            forecast = np.concatenate([oob_forecast, oos_forecast], axis=0)
            forecast = pd.DataFrame(forecast, columns=["forecast"],
                                   index=pred_idx)
            forecasts[keys] = forecast
            
        return {"components":forecasts}
    
    def _validate_dataset(self, data):
        for _, df in data.items():
            try: 
                pd.infer_freq(df.index)
            except TypeError:
                raise ValueError("Pandas cannot infer frequency from the provided "+
                                 "data's index. Please make sure that your data "+
                                 "has a valid datetime index. See panda's " +
                                 "documentation for more information.")    
        return data



from setuptools import setup, find_packages

setup(
    name='aionx',
    version='1.0.7',
    author='Mikael Frenette',
    author_email="<mik.frenette@gmail.com>",
    description='Library for designing Machine Learning models for time series forecast',
    packages=find_packages(),
    py_modules=['aionx'],
    install_requires=[
        "numpy",
        "pandas==2.0.0",
        "tensorflow",
        "scikit-learn",
    ],
)



"""
aionx utils.py is a file for storing/dumping utility classes.

The module contains:

    - The 'add_trends' function:
        A function for adding trends to the provided dataframe.
        
    - The 'WindowEstimationHandler' class:
        generates expanding windows from the provided pandas DataFrame.

"""

# Author: Mikael Frenette (mik.frenette@gmail.com)
# Created: August 2023


import pandas as pd
import numpy as np
from datetime import datetime
from typing import Union
from aionx.wrappers import downstream

@downstream(
    pd.DataFrame, dtype="float32"
)
def add_trends(data: pd.DataFrame, trends: int) -> pd.DataFrame:
    """
    DESCRIPTION
    ---------------------------------------------------------------------------
    Add trend columns to a given pandas DataFrame.
    ---------------------------------------------------------------------------

    PARAMETERS
    ----------
    data   : The original dataframe.

    trends : The number of trend columns to add.

    RETURNS
    -------
        The dataframe with added trend columns.
    """

    try:
        # Create trend columns as separate dataframes with values ranging
        #from 0 to len(data)
        ntrends = [pd.DataFrame(np.arange(0, len(data), 1),
                                columns=[f"trend_{i}"]) for i in range(trends)]

        # Concatenate the trend columns horizontally
        ntrends = pd.concat(ntrends, axis=1)

        # Set the index of ntrends to match the index of the original dataset
        ntrends.index = data.index

        # Concatenate the original dataset and the trend columns vertically
        data = pd.concat([data, ntrends], axis=1)
    except Exception as e:
        print(e)
    
    return data

class ExpandingWindowGenerator:
    """
    DESCRIPTION
    ---------------------------------------------------------------------------
    This class is designed for managing the estimation windows in time series
    data. it generates expanding windows for the training as well as providing
    the remaining out-of-sample data.
    ---------------------------------------------------------------------------

    PARAMETERS
    ----------
    data            : The input time series data.
    
    expanding_start : The index where out-of-sample (OOS) data begins.
    
    timesteps       : The number of time steps in each window.
    
    last_window     : The index of the last window to consider.
    
    verbose         : the level of verbosity

    EXAMPLE
    -------
    # Initialize a WindowEstimationHandler
    expanding_window = WindowEstimationHandler(your_data,
                                        expanding_start=2007-01-01,
                                        timesteps = 12,
                                        last_window="2020-01-01")
    for step, train, oos in expanding_window:
        ...
        your code here
        ...
        
    """
    def __init__(self, 
                 data:pd.DataFrame,
                 expanding_start:Union[str, datetime],
                 timesteps:int,
                 last_window:Union[str, datetime],
                 verbose:int=0) -> None:

        self.expanding_start=expanding_start
        self.timesteps = timesteps
        self.last_window = last_window
        self.verbose=verbose
        self._it = 0    
        
        self.data = data
        if not isinstance(self.data, tuple):
            self.data = (self.data,)

        if isinstance(self.last_window, str):
            self.last_window = datetime.strptime(self.last_window, "%Y-%m-%d")
        if isinstance(self.expanding_start, str):
            self.expanding_start = datetime.strptime(self.expanding_start, "%Y-%m-%d")
        self.starting_idx = len(self.data[0].loc[
            :self.data[0].index[self.data[0].index < self.expanding_start].max()
        ])  
        self.max_it = len(self.data[0].loc[self.expanding_start:self.last_window]) / self.timesteps
        
    def __iter__(self):
        return self
    
    def __update_state(self):
        if self._it >= self.max_it:
            raise StopIteration
        else:
            self._start_point = 0
            self._end_point = self.starting_idx+(self.timesteps*(self._it))
        self._it += 1
            
    def __next__(self):           
        self.__update_state()
        train_slice = [self.data[i].iloc[
            self._start_point:self._end_point] for i in range(len(self.data))]
        
        poos_slice  = [
            self.data[i].iloc[self._end_point:] for i in range(len(self.data))]
        
        if len(train_slice) == 1:
            train_slice=train_slice[0]
        if len(poos_slice) == 1:
            poos_slice=poos_slice[0]
        
        if self.verbose == 1:
            print(f"Expanding[{self._it}/{int(self.max_it)}]")
        if self.verbose > 1:
            print(f"Expanding[{self._it}/{int(self.max_it)}] : \n", 
                    f"    Estimation start: {train_slice[0].index[0].strftime('%Y-%m-%d')}",
                    f" - Estimation end: {train_slice[0].index[-1].strftime('%Y-%m-%d')}") 
        return self._it, train_slice, poos_slice



"""
losses.py is a module designed to store loss functions that inherit
from keras.losses.loss. The classes defined in this module are exclusively
useful within the TensorFlow/Keras ecosystem.

The module contains:

    - The 'GaussianLogLikelihood' class: This class computes the negative of the
      log-likelihood on a provided batch and is used as a keras.losses.loss.
      
"""

# Author: Mikael Frenette (mik.frenette@gmail.com)
# Created: September 2023

import tensorflow as tf
import numpy as np
from typing import Union

class GaussianLogLikelihood(tf.keras.losses.Loss):
    """
    @Author: Mikael Frenette
    
    DESCRIPTION:
    ---------------------------------------------------------------------------
    This function computes the negative log-likelihood loss between the true
    and predicted values using a Gaussian distribution. The loss is calculated
    as the negative sum of the log-probability of the true values given the
    predicted mean and volatility.
    
    Alternatively, you can use the function provided by TensorFlow:
    tf.reduce_sum(tfp.distributions.Normal(loc=μ, scale=σ).log_prob(y))
    
    In this implementation, we have customized the loss function to gain more
    control over its behavior. Specifically, we have added a clip operation
    on the volatility value to avoid numerical instability. 
    This helps with early stopping, as the convergence of the validation loss
    is expected to be more stable.
    ---------------------------------------------------------------------------
    
    PARAMETERS:
    -----------
    volatility_clip : The minimum value for the volatility. Default is 0.05.
        
     See keras.losses.loss documentation for more details. 
    ---------------------------------------------------------------------------
    """
    def __init__(self, volatility_clip:float=0.05, **kwargs):
        
        super().__init__(**kwargs) # inherits from tensorflow.losses.Loss
        
        # the minimum value for the volatility
        self.clip = volatility_clip
    

    @tf.function
    def call(self, y_true: Union[tf.Tensor, np.ndarray],
                    y_pred: Union[tf.Tensor, np.ndarray]):
        """
        PARAMETERS
        ----------
        y_true : The true target values, with a shape of (batch_size, num_outputs)
                 or (batch_size, num_timesteps, num_outputs)
                 in the case of sequence-to-sequence modeling.
                
        y_pred : The predicted target values, consisting of mean and volatility,
                 each with a shape of (batch_size, num_outputs*2) or
                 (batch_size, num_timesteps, num_outputs*2)
                 in the case of sequence-to-sequence modeling.
    
        RETURNS
            loss : The computed negative log-likelihood loss.
        """
        
        mu, sigma = tf.split(y_pred,
                             num_or_size_splits=2,
                             axis=-1)

        mu = tf.cast(tf.squeeze(mu), tf.float32)  # Extract the predicted mean value.
        sigma = tf.cast(tf.squeeze(sigma), tf.float32)  # Extract the predicted volatility value.        
        y = tf.cast(tf.squeeze(y_true), tf.float32)
        # Squeeze the tensors to remove any extra dimensions.
        
        # Clip the volatility to prevent numerical instability.
        sigma = tf.where(sigma > self.clip, sigma, self.clip)  
        
        # Compute the loss using the log-likelihood of the normal distribution.
        loss = tf.reduce_mean(
            tf.math.square(y - mu) / tf.math.square(sigma) + tf.math.log(tf.math.square(sigma))
            )
        
        #loss = -tf.reduce_sum(tfp.distributions.Normal(loc=μ, scale=σ).log_prob(y))
                                                    
        # Return the computed loss as a double-precision float
        return tf.cast(loss, tf.double)  

class InflationMSE(tf.keras.losses.Loss):
    """
    @Author: Mikael Frenette
    
    DESCRIPTION:
    ---------------------------------------------------------------------------
    This function computes the mean squared error (MSE) for the components
    sum with the target values
    ---------------------------------------------------------------------------
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs) # inherits from tensorflow.losses.Loss
        self.MSE = tf.keras.losses.MeanSquaredError()
        
    @tf.function
    def call(self, y_true: Union[tf.Tensor, np.ndarray],
                    y_pred: Union[tf.Tensor, np.ndarray]):
        """
        PARAMETERS
        ----------
        y_true : The true target values, with a shape of (batch_size, num_outputs)
                 or (batch_size, num_timesteps, num_outputs)
                 in the case of sequence-to-sequence modeling.
                
        y_pred : The predicted target values, consisting of mean and volatility,
                 each with a shape of (batch_size, num_outputs*2) or
                 (batch_size, num_timesteps, num_outputs*2)
                 in the case of sequence-to-sequence modeling.
        """
        
        y_pred = tf.reduce_sum(y_pred, keepdims=True, axis=-1)
        loss = self.MSE(y_true, y_pred)
        return tf.cast(loss, tf.double)
